#    Copyright (C) 2008 Niels Richard Hansen. 
# 
#    This file is part of StemSearch.
#
#    StemSearch is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 3 of the License, or
#    (at your option) any later version.
#
#    StemSearch is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program. If not, see <http://www.gnu.org/licenses/>.


require(Biostrings)
VERSION = "0.9"

################################################################################
#Parameter IO-functions
################################################################################

read.parameters <- function(file)
{

	parameters <- list()
	temp <- scan(file,comment.char="#",na.string="*",quiet=TRUE)
	temp[is.na(temp)] <- -Inf
	parameters$SScores <- matrix(temp[1:256],ncol=16,byrow=TRUE)
	parameters$LoopPenalties <- temp[257:(length(temp)-22)]
	parameters$NullModel <- matrix(temp[(length(temp)-21):(length(temp)-6)],ncol=4,byrow=TRUE)
	parameters$Norm <- temp[(length(temp)-5):length(temp)]
	return(parameters)
}

write.parameters <- function(file,parameters)
{

write(paste(
"
###################################################################################
# Parameter file, ",file,", for StemSearch. Created ",date()," by 
# the StemSearch R package ",VERSION,".                                                
###################################################################################
#
#Row and column order: aa ca ga ua ac cc gc uc ag cg gg ug au cu gu uu
#Stack and stack-entry transition scores:",sep=""),file)

parameters$SScores[parameters$SScores == -Inf] <- NA

write.table(round(parameters$SScores,4),file,append=TRUE,col.names=FALSE,row.names=FALSE,na="*",sep="\t")


write("\n#Loop penalties\n",file,append=TRUE)
write.table(parameters$LoopPenalties,file,append=TRUE,col.names=FALSE,row.names=FALSE)

write("\n#Null model\n",file,append=TRUE)
write.table(round(parameters$NullModel,4),file,append=TRUE,col.names=FALSE,row.names=FALSE)

write("\n#Normalisation\n",file,append=TRUE)
write.table(round(parameters$Norm,4),file,append=TRUE,col.names=FALSE,row.names=FALSE)

}

###################################################################
#Estimating transitions in RNA-structures                         #
#with secondary structures for the RNA molecules in file "foo"    #
#call stack.trans.prop(foo) for the stack transitions.		  #
#Requires library Biostrings.					  #
###################################################################

bondpairs <- function(instruct)		#structure in backet notation 
{
	struct <- BString(instruct)
		
	bonds <- list()
	lstruct <- length(struct)
	fiveprime <- 1:lstruct
	indx <- 0
	for(i in 1:lstruct)
	{
		if(letter(struct,i) == "(") 
		{
			indx <- indx +1
			fiveprime[indx] <- i
		}
		else if(letter(struct,i) == ")")
		{
			if(indx == 0) {break}
			bonds <- c(bonds,list(c(fiveprime[indx],i)))
			indx <- indx -1
		}
	}
	return(bonds)
}

trans.freq <- function(inseq,pairs)
{
	seq <- BString(inseq)
	

	lev <- c("aa","ca","ga","ua","ac","cc","gc","uc","ag","cg","gg","ug","au","cu","gu","uu")

	inner <- factor(paste(letter(seq,pairs[[1]][1]+1),letter(seq,pairs[[1]][2]-1),sep=""),levels=lev)
	outer <- factor(paste(letter(seq,pairs[[1]][1]),letter(seq,pairs[[1]][2]),sep=""),levels=lev)	

	pairlist <- data.frame(inner,outer)
		
	for(i in 2:length(pairs))
	{
		inner <- factor(paste(letter(seq,pairs[[i]][1]+1),letter(seq,pairs[[i]][2]-1),sep=""),levels=lev)
		outer <- factor(paste(letter(seq,pairs[[i]][1]),letter(seq,pairs[[i]][2]),sep=""),levels=lev)	

		pairlist <- rbind(pairlist,data.frame(inner,outer)) 
	}
	return(table(pairlist))
}
		
readRNAstruct <- function(file)
{
	temp <- scan(file,what="character",sep="\n")
	RNAstruct <- list()	
	tmp <- list()

	for(i in (0:(length(temp)-1)%/% 3 +1))
	{

		j <- (i-1)*3 + 1
		tmp$desc <- temp[j]
		tmp$seq <- chartr("t","u",tolower(temp[j+1]))
		tmp$struct <- substr(temp[j+2],1,length(BString(tmp$seq)))
		RNAstruct[[i]] <- tmp
	}
	return(RNAstruct)
}

stack.trans.prob <- function(file)
{
	temp <- readRNAstruct(file)	
	tf <- trans.freq(temp[[1]]$seq,bondpairs(temp[[1]]$struct))	

	for(i in 2:length(temp))
	{
		tf <- tf + trans.freq(temp[[i]]$seq,bondpairs(temp[[i]]$struct))
	}

	dimen <- dim(tf)
	transprob <- tf/(matrix(tf %*% rep(1,dimen[2]),nrow=dimen[1],ncol=dimen[2]))
	return(transprob)
}


###################################################################
#Estimating transitions for null model                            #
#With sequence data in  FASTA files "foo1", ..., "fooN"  call     #
#null.trans.prop(c("foo1",...,"fooN")) for the transitions	  #
#probabilities. The function returns a list with the first element# 
#being the totally estimated transition probabilities and the     #
#following entries being separate estimates for each file.        # 
#Requires library Biostrings.					  #
###################################################################



count.trans <- function(file)
{
	patterns <- c("aa","ca","ga","ta","ac","cc","gc","tc","ag","cg","gg","tg","at","ct","gt","tt")

	counts <- numeric(length(patterns))
	genome.seq <- BString(tolower(readFASTA(file)[[1]]$seq))
	for(j in 1:length(patterns))
	{
		counts[j] <- countPattern(patterns[j],genome.seq)
	}
	return(counts)
}


null.trans.prob <- function(file,names=NULL){

if(is.null(names) | length(files) != length(names)) names <- file
trans.prob  <- list(matrix(0,nrow=4,ncol=4,dimnames=list(c("a","c","g","t"),c("a","c","g","t"))))

for(i in 1:length(file))
{
	co <- count.patterns(file[i])
	tmp <- matrix(co,nrow=4,dimnames=list(c("a","c","g","t"),c("a","c","g","t")))
	trans.prob[[1]] <- trans.prob[[1]] + tmp
	tmp <-  tmp / matrix(tmp %*% rep(1,4),nrow=4,ncol=4)
	trans.prob <- c(trans.prob,list(tmp))
}
	tmp <- trans.prob[[1]]
	trans.prob[[1]] <- tmp / matrix(tmp %*% rep(1,4),nrow=4,ncol=4)
	names(trans.prob) <- c("all",names)

	return(trans.prob)
}

###################################################################
#Compute dinucleotide scores based on a matrix of null transition #
#probabilities and stack transition probabilities.                #
###################################################################  

stack.scores <- function(tp,stack.tp){

PmI <- t(tp - diag(c(1,1,1,1)))
PmI[4,] <- c(1,1,1,1)
tp.pi <- solve(PmI,c(0,0,0,1))			#Computes invariant distribution
rev.tp <- ((tp.pi^{-1}) %o% tp.pi) * t(tp)	#Computes reversed transition probabilities

patterns <- c("aa","ca","ga","ua","ac","cc","gc","uc","ag","cg","gg","ug","au","cu","gu","uu")
s <- matrix(nrow=16,ncol=16,dimnames=list(patterns,patterns))

	for(x in 1:4)
	{
		for(y in 1:4)
		{
			for(v in 1:4)
			{ 
				for(w in 1:4)
				{
					i <- x + 4*(y-1)
					j <- v + 4*(w-1)
					s[i,j] <- log(stack.tp[i,j]/(rev.tp[y,x]*tp[v,w]))
			
	}}}}
	return(s)
}

update.stack.scores <- function(parameters,stack.tp){		

	temp.parameters <- parameters
	temp.parameters$SScores <- stack.scores(parameters$NullModel,stack.tp)
	return(temp.parameters)
}


###################################################################
#Estimation of normalization parameters.			  #
###################################################################  

hill.cor <- function(overshoots,par.init=c(1,0,1)){

#par[1] = lambda, par[2] = alpha,par[3] = gamma#

mll <- function(par=c(1,0,1)){
if(!all(c(par[2] <= par[1]*par[3],par >= 0))) {
result <- Inf
} else {
s1 <- (1+overshoots/par[3])
s2 <- log(s1)
result <- sum(-(par[2]-1)*s2 - log(par[1]*s1-par[2]/par[3]) + par[1]*overshoots)
}
return(result)
}

dmll <- function(par=c(1,0,1)){
if(!all(c(par[2] <= par[1]*par[3],par >= 0))){
	result <- c(0,0,0)
} else {
s1 <- (1+overshoots/par[3])
s2 <- log(s1)
s3 <- overshoots + par[3]
s4 <- par[3]*s3
dlambda <- sum(overshoots - s1/(par[1]*s1-par[2]/par[3]))
dalpha <- sum(1/(par[1]*s3 -par[2]) -s2)
dgamma <- sum((par[2]-1)*overshoots/s4 + (par[1]*overshoots-par[3])/(par[1]*s4 -par[2]*par[3]))
result <- c(dlambda,dalpha,dgamma)
}
return(result)
}

par.hat <- optim(par.init,mll)
#par.hat <- optim(par.init,mll,dmll,method="BFGS")
return(par.hat)
}

est.norm.lp <- function(parameters,n=100000,N=200,threshold=3,L=200,file=NULL)
{
	out.file <- file
	if(is.null(file)){

		out.file <- "myoutputfile"
		if(.Platform$OS.type == "windows"){
		command <- paste(Sys.getenv("COMSPEC"), 
				 "/c", 
				 "StemSearch.exe -p Rtemp.par -id -o s -t", threshold, 
				 "-L", format(L,scientific=FALSE),
				 "-I 0", format(n,scientific=FALSE), 
				 "-nr", format(N,scientific=FALSE), 
				 "> ", out.file)
		} else if(.Platform$OS.type == "unix"){
		command <- paste("StemSearch -p Rtemp.par -id -o s -t", threshold, 
				 "-L", format(L,scientific=FALSE),
				 "-I 0", format(n,scientific=FALSE), 
				 "-nr", format(N,scientific=FALSE), 
				 "> ", out.file)
		}	
		

		temp.parameters <- parameters
		temp.parameters$Norm[c(1,2,3)] <- c(1,0,1)
		write.parameters("Rtemp.par",temp.parameters)
	
		system(command, invisible=TRUE)
	}

	scores <- sort(read.table(out.file,header=TRUE)$score,decreasing=TRUE)	#data should be ordered, but for safety if user input file is not ordered

	l <- length(scores)						#Should be N, but can be smaller if t is too large
	if(l < N) print(paste(c("Warning: Hill estimator reduced to POT estimator with threshold ",threshold," and N_t equal to ",l)))

	param <- 1/mean(scores - scores[l])				#Estimate lambda using the Hill estimator (POT if l < N)
	param[2] <- param[1]*scores[l] + log(l/n)   	          	#Estimate log(K)

	#if(is.null(file)) unlink(c("Rtemp.par",out.file)) 		#Tidy up

	temp.parameters$Norm[c(1,2)] <- param

	return(temp.parameters)	
}


est.norm.nolp <- function(parameters,n=10000,N=1000,threshold=3,L=2000,file=NULL)
{
	out.file <- file
	if(is.null(file)){

		out.file <- "myoutputfile"
		if(.Platform$OS.type == "windows"){
		command <- paste(Sys.getenv("COMSPEC"), 
				 "/c", 
				 "StemSearch.exe -p Rtemp.par -nolp -id -o s -t", threshold, 
				 "-L", format(L,scientific=FALSE),
				 "-I 0", format(n,scientific=FALSE), 
				 "-nr", format(N,scientific=FALSE), 
				 "> ", out.file)
		} else if(.Platform$OS.type == "unix"){
		command <- paste("StemSearch -p Rtemp.par -nolp -id -o s -t", threshold, 
				 "-L", format(L,scientific=FALSE),
				 "-I 0", format(n,scientific=FALSE), 
				 "-nr", format(N,scientific=FALSE), 
				 "> ", out.file)
		}	


		temp.parameters <- parameters
		temp.parameters$Norm[c(4,5,6)] <- c(1,0,1)
		write.parameters("Rtemp.par",temp.parameters)
	
		system(command, invisible=TRUE)
	}

	scores <- sort(read.table(out.file,header=TRUE)$score,decreasing=TRUE)	#data should be ordered, but for safety if user input file is not ordered

	l <- length(scores)						#Should be N, but can be smaller if t is to low
	if(l < N) print(paste(c("Warning: Hill estimator reduced to POT estimator with threshold ",threshold," and N_t equal to ",l)))

	param <- 1/mean(scores - scores[l])				#Estimate lambda using the Hill estimator (POT if l < N)
	param[2] <- param[1]*scores[l] + log(l/((L-1)*(n-L/2)))        	#Estimate log(K)

	if(is.null(file)) unlink(c("Rtemp.par",out.file)) 		#Tidy up

	temp.parameters$Norm[c(4,5)] <- param

	return(temp.parameters)	
}

