/*  Copyright (C) 2008 Niels Richard Hansen. 

    This file is part of StemSearch.

    StemSearch is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 3 of the License, or
    (at your option) any later version.

    StemSearch is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program. If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef STRUCTSEARCH_INC_GUARD
#define STRUCTSEARCH_INC_GUARD
#define VERSION "0.9"

#include "../matrix/matrix.h"
#include "../genome/genome.h"
#include <string>
#include <vector>
#include <deque>

class Island
{
private:
	double Score;			//Normalized
	double Evalue;	                //Expectation value - based on Score and length of 
	string GenomeName;              //Identifier for the genomic sequence where the island is located
	long BandSize;
	long IslandNr;
	string Sequence;
	string Structure3;
	string Structure5;
	vector<long> Anchor;
	vector<long> Pos;
public:
	Island();
	double& score();
	double score() const;
	double& evalue();
	string& genome();
	long& bandsize();
	long& islandnr();
	string& sequence();
	string& structure3();
	string& structure5();
	void anchor(long,long);
	void pos(long,long);
	string pretty_print();
	string pretty_print_minimal();
	string pretty_print_nl();
	string pretty_print_gff(long);
	long operator[](long);
	long operator()(long);
	
};	

class StructSearch
{
private:
	bool eos;
	Genome G;
	RectMatrix<double> SScore;		        //Stack pair transition scores
	RectMatrix<double> NullTP;		        //Transition probabilities under a first order Markov model
	RectMatrix<int> StackMask;		        //Masking matrices determined from parameter file	
	vector<double> HPLoopPen;
	
	long start,end,band,L,l;			//The width of the diagonal band in the score, island and traceback matrices and genome size
	double threshold, Le, Li;                       //Threshold for (island) scores, loop extension and loop initialisation penalty
	int SIT;					// 1=ST, 2=SIT, 3=EST, 4 = ESIT
	UpBandMatrix<double> SS,LS;	 		//Stack Score and Loop Score matrices
	UpBandMatrix<long> ST,LT;			//Traceback matrices
	UpBandMatrix<long> SI,LI;			//Island matrices
	RectMatrix<double> ESS,ELS;			//Memory Efficient Stack Score and Loop Score matrices
	RectMatrix<long> EST,ELT,ESI,ELI;		//Memory Efficient Traceback and Island matrices
			
	double lambda, logK, H;                         //Normalization parameters (use of H currently not implemented)
	bool warnings, hplooppen;
	
	bool store_random_seq;
	
	deque<Island> seqstr;				//High Scoring Sequence segment and addtional information 
							//declumped according to the island matrix
	vector<long> MaxPosi, MaxPosj;
	
	void Etraceback(Island&);
	void read_par(const char*);
	void error(char*) const;
	void IgnoreLines(istream&);
	string defaultpar(const char*);
public:
	StructSearch(long band_size, double threshold, const char* parameter_filename, bool Lmethod =true , bool =false, long =1, long = 0);
	void read(istream&,char='\n');
	bool done();
	void print_score_matrices();
	void computeST();
	void computeSIT();
	void computeEST();
	void computeESIT();
	void saveS();
	void saveI();
	void traceback();
	void traceback(Island&);
	void greedystr();
	string pretty_print(long=0,int=0);
	vector<double> excesses(long =0);
	void generate(long);
};

ostream& operator<<(ostream&,Island&);

ostream& operator<<(ostream&,StructSearch&);

bool str_order(const Island&, const Island&);

#endif  	//STRUCTSEARCH_INC_GUARD
